@@ -73,7 +73,6 @@ struct _SpiceDisplayChannelPrivate {
 #ifdef WIN32
     HDC dc;
 #endif
-    gboolean                    migrate_wait_primary;
 };
 
 G_DEFINE_TYPE(SpiceDisplayChannel, spice_display_channel, SPICE_TYPE_CHANNEL)
@@ -97,10 +96,9 @@ enum {
 static guint signals[SPICE_DISPLAY_LAST_SIGNAL];
 
 static void spice_display_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg);
-static void spice_display_channel_init(SpiceDisplayChannel *channel);
 static void spice_display_channel_up(SpiceChannel *channel);
 
-static void clear_surfaces(SpiceChannel *channel);
+static void clear_surfaces(SpiceChannel *channel, gboolean keep_primary);
 static void clear_streams(SpiceChannel *channel);
 static display_surface *find_surface(SpiceDisplayChannelPrivate *c, int surface_id);
 static gboolean display_stream_render(display_stream *st);
@@ -124,11 +122,7 @@ static void spice_display_channel_dispose(GObject *object)
 
 static void spice_display_channel_finalize(GObject *object)
 {
-    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
-
-    c->migrate_wait_primary = FALSE;
-
-    clear_surfaces(SPICE_CHANNEL(object));
+    clear_surfaces(SPICE_CHANNEL(object), FALSE);
     clear_streams(SPICE_CHANNEL(object));
 
     if (G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize)
@@ -191,13 +185,9 @@ static void spice_display_set_property(GObject      *object,
 /* main or coroutine context */
 static void spice_display_channel_reset(SpiceChannel *channel, gboolean migrating)
 {
-    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
-
     /* palettes, images, and glz_window are cleared in the session */
-
-    c->migrate_wait_primary = migrating;
     clear_streams(channel);
-    clear_surfaces(channel);
+    clear_surfaces(channel, TRUE);
 
     SPICE_CHANNEL_CLASS(spice_display_channel_parent_class)->channel_reset(channel, migrating);
 }
@@ -550,8 +540,8 @@ static pixman_image_t* image_get_lossless(SpiceImageCache *cache, uint64_t id)
 }
 #endif
 
-SpiceCanvas *surfaces_get(SpiceImageSurfaces *surfaces,
-                          uint32_t surface_id)
+static SpiceCanvas *surfaces_get(SpiceImageSurfaces *surfaces,
+                                 uint32_t surface_id)
 {
     SpiceDisplayChannelPrivate *c =
         SPICE_CONTAINEROF(surfaces, SpiceDisplayChannelPrivate, image_surfaces);
@@ -619,10 +609,8 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
         display_surface *primary = find_surface(c, 0);
 
         if (primary) {
-            if (c->migrate_wait_primary &&
-                primary->width == surface->width &&
+            if (primary->width == surface->width &&
                 primary->height == surface->height) {
-                c->migrate_wait_primary = FALSE;
                 SPICE_DEBUG("Reusing existing primary surface");
                 return 0;
             }
@@ -634,8 +622,6 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
         }
 
         SPICE_DEBUG("display: create primary canvas");
-        c->migrate_wait_primary = FALSE;
-
 #ifdef HAVE_SYS_SHM_H
         surface->shmid = shmget(IPC_PRIVATE, surface->size, IPC_CREAT | 0777);
         if (surface->shmid >= 0) {
@@ -706,6 +692,7 @@ static void destroy_canvas(display_surface *surface)
 #ifdef HAVE_SYS_SHM_H
     else {
         shmdt(surface->data);
+        shmctl(surface->shmid, IPC_RMID, 0);
     }
 #endif
     surface->shmid = -1;
@@ -730,7 +717,7 @@ static display_surface *find_surface(SpiceDisplayChannelPrivate *c, int surface_
     return NULL;
 }
 
-static void clear_surfaces(SpiceChannel *channel)
+static void clear_surfaces(SpiceChannel *channel, gboolean keep_primary)
 {
     SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     display_surface *surface;
@@ -740,8 +727,8 @@ static void clear_surfaces(SpiceChannel *channel)
         surface = SPICE_CONTAINEROF(item, display_surface, link);
         item = ring_next(&c->surfaces, item);
 
-        if (c->migrate_wait_primary && surface->primary) {
-            SPICE_DEBUG("Try to keep exisiting primary surface during migration");
+        if (keep_primary && surface->primary) {
+            SPICE_DEBUG("keeping exisiting primary surface, migration or reset");
             continue;
         }
 
@@ -816,11 +803,6 @@ static void display_handle_mode(SpiceChannel *channel, SpiceMsgIn *in)
     surface->size    = surface->height * surface->stride;
     surface->primary = true;
     create_canvas(channel, surface);
-#ifdef HAVE_SYS_SHM_H
-    if (surface->shmid != -1) {
-        shmctl(surface->shmid, IPC_RMID, 0);
-    }
-#endif
 }
 
 /* coroutine context */
