@@ -74,10 +74,7 @@ static void spice_inputs_channel_up(SpiceChannel *channel);
 
 static void spice_inputs_channel_init(SpiceInputsChannel *channel)
 {
-    SpiceInputsChannelPrivate *c;
-
-    c = channel->priv = SPICE_INPUTS_CHANNEL_GET_PRIVATE(channel);
-    memset(c, 0, sizeof(*c));
+    channel->priv = SPICE_INPUTS_CHANNEL_GET_PRIVATE(channel);
 }
 
 static void spice_inputs_get_property(GObject    *object,
@@ -218,12 +215,14 @@ static void send_position(SpiceInputsChannel *channel)
 {
     SpiceMsgOut *msg;
 
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
+
     msg = mouse_position(channel);
     if (!msg) /* if no motion */
         return;
 
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /* main context */
@@ -231,12 +230,14 @@ static void send_motion(SpiceInputsChannel *channel)
 {
     SpiceMsgOut *msg;
 
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
+
     msg = mouse_motion(channel);
     if (!msg) /* if no motion */
         return;
 
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /* coroutine context */
@@ -270,13 +271,11 @@ static void inputs_handle_ack(SpiceChannel *channel, SpiceMsgIn *in)
     msg = mouse_motion(SPICE_INPUTS_CHANNEL(channel));
     if (msg) { /* if no motion, msg == NULL */
         spice_msg_out_send_internal(msg);
-        spice_msg_out_unref(msg);
     }
 
     msg = mouse_position(SPICE_INPUTS_CHANNEL(channel));
     if (msg) {
         spice_msg_out_send_internal(msg);
-        spice_msg_out_unref(msg);
     }
 }
 
@@ -323,6 +322,9 @@ void spice_inputs_motion(SpiceInputsChannel *channel, gint dx, gint dy,
     if (SPICE_CHANNEL(channel)->priv->state != SPICE_CHANNEL_STATE_READY)
         return;
 
+    if (dx == 0 && dy == 0)
+        return;
+
     c = channel->priv;
     c->bs  = button_state;
     c->dx += dx;
@@ -385,6 +387,8 @@ void spice_inputs_button_press(SpiceInputsChannel *channel, gint button,
 
     if (SPICE_CHANNEL(channel)->priv->state != SPICE_CHANNEL_STATE_READY)
         return;
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
 
     c = channel->priv;
     switch (button) {
@@ -409,7 +413,6 @@ void spice_inputs_button_press(SpiceInputsChannel *channel, gint button,
     press.buttons_state = button_state;
     msg->marshallers->msgc_inputs_mouse_press(msg->marshaller, &press);
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /**
@@ -431,6 +434,8 @@ void spice_inputs_button_release(SpiceInputsChannel *channel, gint button,
 
     if (SPICE_CHANNEL(channel)->priv->state != SPICE_CHANNEL_STATE_READY)
         return;
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
 
     c = channel->priv;
     switch (button) {
@@ -445,7 +450,7 @@ void spice_inputs_button_release(SpiceInputsChannel *channel, gint button,
         break;
     }
 
-    c->bs  = button_state;
+    c->bs = button_state;
     send_motion(channel);
     send_position(channel);
 
@@ -455,7 +460,6 @@ void spice_inputs_button_release(SpiceInputsChannel *channel, gint button,
     release.buttons_state = button_state;
     msg->marshallers->msgc_inputs_mouse_release(msg->marshaller, &release);
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /**
@@ -474,6 +478,8 @@ void spice_inputs_key_press(SpiceInputsChannel *channel, guint scancode)
     g_return_if_fail(SPICE_CHANNEL(channel)->priv->state != SPICE_CHANNEL_STATE_UNCONNECTED);
     if (SPICE_CHANNEL(channel)->priv->state != SPICE_CHANNEL_STATE_READY)
         return;
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
 
     SPICE_DEBUG("%s: scancode %d", __FUNCTION__, scancode);
     if (scancode < 0x100) {
@@ -486,7 +492,6 @@ void spice_inputs_key_press(SpiceInputsChannel *channel, guint scancode)
                             SPICE_MSGC_INPUTS_KEY_DOWN);
     msg->marshallers->msgc_inputs_key_down(msg->marshaller, &down);
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /**
@@ -505,6 +510,8 @@ void spice_inputs_key_release(SpiceInputsChannel *channel, guint scancode)
     g_return_if_fail(SPICE_CHANNEL(channel)->priv->state != SPICE_CHANNEL_STATE_UNCONNECTED);
     if (SPICE_CHANNEL(channel)->priv->state != SPICE_CHANNEL_STATE_READY)
         return;
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
 
     SPICE_DEBUG("%s: scancode %d", __FUNCTION__, scancode);
     if (scancode < 0x100) {
@@ -517,7 +524,6 @@ void spice_inputs_key_release(SpiceInputsChannel *channel, guint scancode)
                             SPICE_MSGC_INPUTS_KEY_UP);
     msg->marshallers->msgc_inputs_key_up(msg->marshaller, &up);
     spice_msg_out_send(msg);
-    spice_msg_out_unref(msg);
 }
 
 /* main or coroutine context */
@@ -555,12 +561,14 @@ void spice_inputs_set_key_locks(SpiceInputsChannel *channel, guint locks)
 {
     SpiceMsgOut *msg;
 
+    if (spice_channel_get_read_only(SPICE_CHANNEL(channel)))
+        return;
+
     msg = set_key_locks(channel, locks);
     if (!msg) /* you can set_key_locks() even if the channel is not ready */
         return;
 
     spice_msg_out_send(msg); /* main -> coroutine */
-    spice_msg_out_unref(msg);
 }
 
 /* coroutine context */
@@ -569,7 +577,9 @@ static void spice_inputs_channel_up(SpiceChannel *channel)
     SpiceInputsChannelPrivate *c = SPICE_INPUTS_CHANNEL(channel)->priv;
     SpiceMsgOut *msg;
 
+    if (spice_channel_get_read_only(channel))
+        return;
+
     msg = set_key_locks(SPICE_INPUTS_CHANNEL(channel), c->locks);
     spice_msg_out_send_internal(msg);
-    spice_msg_out_unref(msg);
 }
