@@ -47,7 +47,6 @@ typedef struct _SpiceAudioPrivate SpiceAudioPrivate;
 
 /**
  * SpiceAudio:
- * @parent: Parent instance.
  *
  * The #SpiceAudio struct is opaque and should not be accessed directly.
  */
@@ -73,10 +72,13 @@ struct _SpiceAudioClass {
 
 GType spice_audio_get_type(void);
 
-SpiceAudio* spice_audio_new(SpiceSession *session, GMainContext *context, const char *name);
-
 SpiceAudio* spice_audio_get(SpiceSession *session, GMainContext *context);
 
+#ifndef SPICE_DISABLE_DEPRECATED
+SPICE_DEPRECATED_FOR(spice_audio_get)
+SpiceAudio* spice_audio_new(SpiceSession *session, GMainContext *context, const char *name);
+#endif
+
 G_END_DECLS
 
 #endif /* __SPICE_CLIENT_AUDIO_H__ */
