@@ -29,14 +29,13 @@
 #include <sasl/sasl.h>
 #endif
 
+#include "spice-util-priv.h"
 #include "coroutine.h"
 #include "gio-coroutine.h"
 
-/* common/ */
-#include "marshallers.h"
-#include "demarshallers.h"
-
-#include "ssl_verify.h"
+#include "common/client_marshallers.h"
+#include "common/client_demarshallers.h"
+#include "common/ssl_verify.h"
 
 G_BEGIN_DECLS
 
@@ -101,7 +100,7 @@ struct _SpiceChannelPrivate {
 
     GQueue                      xmit_queue;
     gboolean                    xmit_queue_blocked;
-    GStaticMutex                xmit_queue_lock;
+    STATIC_MUTEX                xmit_queue_lock;
     guint                       xmit_queue_wakeup_id;
 
     char                        name[16];
@@ -157,6 +156,7 @@ void spice_channel_up(SpiceChannel *channel);
 void spice_channel_wakeup(SpiceChannel *channel, gboolean cancel);
 
 SpiceSession* spice_channel_get_session(SpiceChannel *channel);
+enum spice_channel_state spice_channel_get_state(SpiceChannel *channel);
 
 /* coroutine context */
 typedef void (*handler_msg_in)(SpiceChannel *channel, SpiceMsgIn *msg, gpointer data);
