@@ -21,6 +21,7 @@
 #include "spice-channel-priv.h"
 #include "spice-session-priv.h"
 #include "spice-marshal.h"
+#include "bio-gsocket.h"
 
 #include <openssl/rsa.h>
 #include <openssl/evp.h>
@@ -109,7 +110,7 @@ static void spice_channel_init(SpiceChannel *channel)
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_PROTOCOL_AUTH_SELECTION);
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_MINI_HEADER);
     g_queue_init(&c->xmit_queue);
-    g_static_mutex_init(&c->xmit_queue_lock);
+    STATIC_MUTEX_INIT(c->xmit_queue_lock);
 }
 
 static void spice_channel_constructed(GObject *gobject)
@@ -159,7 +160,9 @@ static void spice_channel_finalize(GObject *gobject)
 
     SPICE_DEBUG("%s: %s %p", c->name, __FUNCTION__, gobject);
 
-    g_idle_remove_by_data (gobject);
+    g_idle_remove_by_data(gobject);
+
+    STATIC_MUTEX_CLEAR(c->xmit_queue_lock);
 
     if (c->caps)
         g_array_free(c->caps, TRUE);
@@ -529,7 +532,7 @@ void *spice_msg_in_raw(SpiceMsgIn *in, int *len)
     return in->data;
 }
 
-static void hexdump(char *prefix, unsigned char *data, int len)
+static void hexdump(const char *prefix, unsigned char *data, int len)
 {
     int i;
 
@@ -664,9 +667,9 @@ static gboolean spice_channel_idle_wakeup(gpointer user_data)
      *   5) xmit_queue_wakeup_id now says there is a wakeup pending which is
      *      false
      */
-    g_static_mutex_lock(&c->xmit_queue_lock);
+    STATIC_MUTEX_LOCK(c->xmit_queue_lock);
     c->xmit_queue_wakeup_id = 0;
-    g_static_mutex_unlock(&c->xmit_queue_lock);
+    STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
 
     spice_channel_wakeup(channel, FALSE);
 
@@ -677,27 +680,32 @@ static gboolean spice_channel_idle_wakeup(gpointer user_data)
 G_GNUC_INTERNAL
 void spice_msg_out_send(SpiceMsgOut *out)
 {
+    gboolean was_empty;
+
     g_return_if_fail(out != NULL);
     g_return_if_fail(out->channel != NULL);
 
-    g_static_mutex_lock(&out->channel->priv->xmit_queue_lock);
-    if (!out->channel->priv->xmit_queue_blocked) {
-        gboolean was_empty;
-
-        was_empty = g_queue_is_empty(&out->channel->priv->xmit_queue);
-        g_queue_push_tail(&out->channel->priv->xmit_queue, out);
-
-        /* One wakeup is enough to empty the entire queue -> only do a wakeup
-           if the queue was empty, and there isn't one pending already. */
-        if (was_empty && !out->channel->priv->xmit_queue_wakeup_id) {
-            out->channel->priv->xmit_queue_wakeup_id =
-                /* Use g_timeout_add_full so that can specify the priority */
-                g_timeout_add_full(G_PRIORITY_HIGH, 0,
-                                   spice_channel_idle_wakeup,
-                                   out->channel, NULL);
-        }
+    STATIC_MUTEX_LOCK(out->channel->priv->xmit_queue_lock);
+    if (out->channel->priv->xmit_queue_blocked) {
+        g_warning("message queue is blocked, dropping message");
+        goto end;
     }
-    g_static_mutex_unlock(&out->channel->priv->xmit_queue_lock);
+
+    was_empty = g_queue_is_empty(&out->channel->priv->xmit_queue);
+    g_queue_push_tail(&out->channel->priv->xmit_queue, out);
+
+    /* One wakeup is enough to empty the entire queue -> only do a wakeup
+       if the queue was empty, and there isn't one pending already. */
+    if (was_empty && !out->channel->priv->xmit_queue_wakeup_id) {
+        out->channel->priv->xmit_queue_wakeup_id =
+            /* Use g_timeout_add_full so that can specify the priority */
+            g_timeout_add_full(G_PRIORITY_HIGH, 0,
+                               spice_channel_idle_wakeup,
+                               out->channel, NULL);
+    }
+
+end:
+    STATIC_MUTEX_UNLOCK(out->channel->priv->xmit_queue_lock);
 }
 
 /* coroutine context */
@@ -1055,14 +1063,14 @@ static void spice_channel_recv_auth(SpiceChannel *channel)
 
     rc = spice_channel_read(channel, &link_res, sizeof(link_res));
     if (rc != sizeof(link_res)) {
-        g_critical("incomplete auth reply (%d/%" G_GSIZE_FORMAT ")",
-                   rc, sizeof(link_res));
+        SPICE_DEBUG("incomplete auth reply (%d/%" G_GSIZE_FORMAT ")",
+                    rc, sizeof(link_res));
         emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_LINK);
         return;
     }
 
     if (link_res != SPICE_LINK_ERR_OK) {
-        g_critical("link result: reply %d", link_res);
+        SPICE_DEBUG("link result: reply %d", link_res);
         emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_AUTH);
         return;
     }
@@ -1431,7 +1439,7 @@ restart:
     /* Need to gather some credentials from the client */
     if (err == SASL_INTERACT) {
         if (!spice_channel_gather_sasl_credentials(channel, interact)) {
-            g_critical("Failed to collect auth credentials");
+            SPICE_DEBUG("Failed to collect auth credentials");
             goto error;
         }
         goto restart;
@@ -1512,7 +1520,7 @@ restart:
         if (err == SASL_INTERACT) {
             if (!spice_channel_gather_sasl_credentials(channel,
                                                        interact)) {
-                g_critical("%s", "Failed to collect auth credentials");
+                SPICE_DEBUG("%s", "Failed to collect auth credentials");
                 goto error;
             }
             goto restep;
@@ -1612,8 +1620,7 @@ complete:
 error:
     if (saslconn)
         sasl_dispose(&saslconn);
-    if (!c->has_error)
-        emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_AUTH);
+    emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_ERROR_AUTH);
     c->has_error = TRUE; /* force disconnect */
     return FALSE;
 }
@@ -1624,6 +1631,7 @@ static void spice_channel_recv_link_msg(SpiceChannel *channel)
 {
     SpiceChannelPrivate *c;
     int rc, num_caps, i;
+    uint32_t *caps;
 
     g_return_if_fail(channel != NULL);
     g_return_if_fail(channel->priv != NULL);
@@ -1661,7 +1669,7 @@ static void spice_channel_recv_link_msg(SpiceChannel *channel)
     /* see original spice/client code: */
     /* g_return_if_fail(c->peer_msg + c->peer_msg->caps_offset * sizeof(uint32_t) > c->peer_msg + c->peer_hdr.size); */
 
-    uint32_t *caps = (uint32_t *)((uint8_t *)c->peer_msg + c->peer_msg->caps_offset);
+    caps = (uint32_t *)((uint8_t *)c->peer_msg + c->peer_msg->caps_offset);
 
     g_array_set_size(c->remote_common_caps, c->peer_msg->num_common_caps);
     for (i = 0; i < c->peer_msg->num_common_caps; i++, caps++) {
@@ -1855,7 +1863,7 @@ const gchar* spice_channel_type_to_string(gint type)
     };
     const char *str = NULL;
 
-    if (type >= 0 && type < sizeof(to_string)) {
+    if (type >= 0 && type < G_N_ELEMENTS(to_string)) {
         str = to_string[type];
     }
 
@@ -1964,9 +1972,9 @@ static void spice_channel_iterate_write(SpiceChannel *channel)
     SpiceMsgOut *out;
 
     do {
-        g_static_mutex_lock(&c->xmit_queue_lock);
+        STATIC_MUTEX_LOCK(c->xmit_queue_lock);
         out = g_queue_pop_head(&c->xmit_queue);
-        g_static_mutex_unlock(&c->xmit_queue_lock);
+        STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
         if (out)
             spice_channel_write_msg(channel, out);
     } while (out);
@@ -2090,12 +2098,13 @@ static void *spice_channel_coroutine(void *data)
             goto cleanup;
         }
 
-	if (!(c->sock = g_socket_new_from_fd(c->fd, NULL))) {
-		SPICE_DEBUG("Failed to open socket from fd %d", c->fd);
-		return FALSE;
-	}
+        if (!(c->sock = g_socket_new_from_fd(c->fd, NULL))) {
+                SPICE_DEBUG("Failed to open socket from fd %d", c->fd);
+                return FALSE;
+        }
 
-	g_socket_set_blocking(c->sock, FALSE);
+        g_socket_set_blocking(c->sock, FALSE);
+        g_socket_set_keepalive(c->sock, TRUE);
         goto connected;
     }
 
@@ -2156,13 +2165,11 @@ reconnect:
             g_critical("SSL_new failed");
             goto cleanup;
         }
-        rc = SSL_set_fd(c->ssl, g_socket_get_fd(c->sock));
-        if (rc <= 0) {
-            g_critical("SSL_set_fd failed");
-            goto cleanup;
-        }
 
 
+        BIO *bio = bio_new_gsocket(c->sock);
+        SSL_set_bio(c->ssl, bio, bio);
+
         {
             guint8 *pubkey;
             guint pubkey_len;
@@ -2193,7 +2200,11 @@ ssl_reconnect:
 connected:
     rc = setsockopt(g_socket_get_fd(c->sock), IPPROTO_TCP, TCP_NODELAY,
                     (const char*)&delay_val, sizeof(delay_val));
-    if (rc != 0) {
+    if ((rc != 0)
+#ifdef ENOTSUP
+        && (errno != ENOTSUP)
+#endif
+        ) {
         g_warning("%s: could not set sockopt TCP_NODELAY: %s", c->name,
                   strerror(errno));
     }
@@ -2302,10 +2313,14 @@ gboolean spice_channel_connect(SpiceChannel *channel)
 /**
  * spice_channel_open_fd:
  * @channel:
- * @fd: a file descriptor (socket)
+ * @fd: a file descriptor (socket) or -1.
+ * request mechanism
  *
  * Connect the channel using @fd socket.
  *
+ * If @fd is -1, a valid fd will be requested later via the
+ * SpiceChannel::open-fd signal.
+ *
  * Returns: %TRUE on success.
  **/
 gboolean spice_channel_open_fd(SpiceChannel *channel, int fd)
@@ -2313,7 +2328,7 @@ gboolean spice_channel_open_fd(SpiceChannel *channel, int fd)
     SpiceChannelPrivate *c = SPICE_CHANNEL_GET_PRIVATE(channel);
 
     g_return_val_if_fail(c != NULL, FALSE);
-    g_return_val_if_fail(fd >= 0, FALSE);
+    g_return_val_if_fail(fd >= -1, FALSE);
 
     c->fd = fd;
 
@@ -2362,7 +2377,7 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
     c->peer_msg = NULL;
     c->peer_pos = 0;
 
-    g_static_mutex_lock(&c->xmit_queue_lock);
+    STATIC_MUTEX_LOCK(c->xmit_queue_lock);
     c->xmit_queue_blocked = TRUE; /* Disallow queuing new messages */
     g_queue_foreach(&c->xmit_queue, (GFunc)spice_msg_out_unref, NULL);
     g_queue_clear(&c->xmit_queue);
@@ -2370,7 +2385,7 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
         g_source_remove(c->xmit_queue_wakeup_id);
         c->xmit_queue_wakeup_id = 0;
     }
-    g_static_mutex_unlock(&c->xmit_queue_lock);
+    STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
 
     g_array_set_size(c->remote_common_caps, 0);
     g_array_set_size(c->remote_caps, 0);
@@ -2400,11 +2415,11 @@ static void channel_disconnect(SpiceChannel *channel)
 
     c->has_error = TRUE; /* break the loop */
 
-    spice_channel_reset(channel, FALSE);
-
     if (c->state == SPICE_CHANNEL_STATE_READY)
         emit_main_context(channel, SPICE_CHANNEL_EVENT, SPICE_CHANNEL_CLOSED);
 
+    spice_channel_reset(channel, FALSE);
+
     g_return_if_fail(SPICE_IS_CHANNEL(channel));
     c->state = SPICE_CHANNEL_STATE_UNCONNECTED;
 }
@@ -2462,7 +2477,7 @@ static gboolean test_capability(GArray *caps, guint32 cap)
 
 /**
  * spice_channel_test_capability:
- * @self:
+ * @channel:
  * @cap:
  *
  * Test availability of remote "channel kind capability".
@@ -2481,7 +2496,7 @@ gboolean spice_channel_test_capability(SpiceChannel *self, guint32 cap)
 
 /**
  * spice_channel_test_common_capability:
- * @self:
+ * @channel:
  * @cap:
  *
  * Test availability of remote "common channel capability".
@@ -2548,6 +2563,15 @@ SpiceSession* spice_channel_get_session(SpiceChannel *channel)
     return channel->priv->session;
 }
 
+G_GNUC_INTERNAL
+enum spice_channel_state spice_channel_get_state(SpiceChannel *channel)
+{
+    g_return_val_if_fail(SPICE_IS_CHANNEL(channel),
+                         SPICE_CHANNEL_STATE_UNCONNECTED);
+
+    return channel->priv->state;
+}
+
 G_GNUC_INTERNAL
 void spice_channel_swap(SpiceChannel *channel, SpiceChannel *swap)
 {
@@ -2560,49 +2584,32 @@ void spice_channel_swap(SpiceChannel *channel, SpiceChannel *swap)
     g_return_if_fail(s->session != NULL);
     g_return_if_fail(s->sock != NULL);
 
-    {
-        GSocket *sock = c->sock;
-        SSL_CTX *ctx = c->ctx;
-        SSL *ssl = c->ssl;
-        SpiceOpenSSLVerify *sslverify = c->sslverify;
-        uint64_t in_serial = c->in_serial;
-        uint64_t out_serial = c->out_serial;
-        gboolean use_mini_header = c->use_mini_header;
-
-        c->sock = s->sock;
-        c->ctx = s->ctx;
-        c->ssl = s->ssl;
-        c->sslverify = s->sslverify;
-        c->in_serial = s->in_serial;
-        c->out_serial = s->out_serial;
-        c->use_mini_header = s->use_mini_header;
-
-        s->sock = sock;
-        s->ctx = ctx;
-        s->ssl = ssl;
-        s->sslverify = sslverify;
-        s->in_serial = in_serial;
-        s->out_serial = out_serial;
-        s->use_mini_header = use_mini_header;
-    }
+#define SWAP(Field) ({                          \
+    typeof (c->Field) Field = c->Field;         \
+    c->Field = s->Field;                        \
+    s->Field = Field;                           \
+})
 
+    /* TODO: split channel in 2 objects: a controller and a swappable
+       state object */
+    SWAP(sock);
+    SWAP(ctx);
+    SWAP(ssl);
+    SWAP(sslverify);
+    SWAP(in_serial);
+    SWAP(out_serial);
+    SWAP(use_mini_header);
+    SWAP(xmit_queue);
+    SWAP(xmit_queue_blocked);
+    SWAP(caps);
+    SWAP(common_caps);
+    SWAP(remote_caps);
+    SWAP(remote_common_caps);
 #if HAVE_SASL
-    {
-        sasl_conn_t *sasl_conn = c->sasl_conn;
-        const char *sasl_decoded = c->sasl_decoded;
-        unsigned int sasl_decoded_length = c->sasl_decoded_length;
-        unsigned int sasl_decoded_offset = c->sasl_decoded_offset;
-
-        c->sasl_conn = s->sasl_conn;
-        c->sasl_decoded = s->sasl_decoded;
-        c->sasl_decoded_length = s->sasl_decoded_length;
-        c->sasl_decoded_offset = s->sasl_decoded_offset;
-
-        s->sasl_conn = sasl_conn;
-        s->sasl_decoded = sasl_decoded;
-        s->sasl_decoded_length = sasl_decoded_length;
-        s->sasl_decoded_offset = sasl_decoded_offset;
-    }
+    SWAP(sasl_conn);
+    SWAP(sasl_decoded);
+    SWAP(sasl_decoded_length);
+    SWAP(sasl_decoded_offset);
 #endif
 }
 
