@@ -28,11 +28,12 @@
 #include "spice-gstaudio.h"
 #include "spice-common.h"
 #include "spice-session.h"
+#include "spice-util.h"
 
 #define SPICE_GSTAUDIO_GET_PRIVATE(obj)                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_GSTAUDIO, spice_gstaudio))
+    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_GSTAUDIO, SpiceGstaudioPrivate))
 
-G_DEFINE_TYPE(SpiceGstAudio, spice_gstaudio, SPICE_TYPE_AUDIO)
+G_DEFINE_TYPE(SpiceGstaudio, spice_gstaudio, SPICE_TYPE_AUDIO)
 
 struct stream {
     GstElement              *pipe;
@@ -42,8 +43,7 @@ struct stream {
     guint                   channels;
 };
 
-struct spice_gstaudio {
-    SpiceSession            *session;
+struct _SpiceGstaudioPrivate {
     SpiceChannel            *pchannel;
     SpiceChannel            *rchannel;
     struct stream           playback;
@@ -53,6 +53,7 @@ struct spice_gstaudio {
 
 static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
                           gpointer data);
+static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel);
 
 static void spice_gstaudio_finalize(GObject *obj)
 {
@@ -80,7 +81,7 @@ void stream_dispose(struct stream *s)
 
 static void spice_gstaudio_dispose(GObject *obj)
 {
-    spice_gstaudio *p;
+    SpiceGstaudioPrivate *p;
     SPICE_DEBUG("%s", __FUNCTION__);
     p = SPICE_GSTAUDIO_GET_PRIVATE(obj);
 
@@ -102,28 +103,28 @@ static void spice_gstaudio_dispose(GObject *obj)
     }
 }
 
-static void spice_gstaudio_init(SpiceGstAudio *pulse)
+static void spice_gstaudio_init(SpiceGstaudio *pulse)
 {
-    spice_gstaudio *p;
-
-    p = pulse->priv = SPICE_GSTAUDIO_GET_PRIVATE(pulse);
-    memset(p, 0, sizeof(*p));
+    pulse->priv = SPICE_GSTAUDIO_GET_PRIVATE(pulse);
 }
 
-static void spice_gstaudio_class_init(SpiceGstAudioClass *klass)
+static void spice_gstaudio_class_init(SpiceGstaudioClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
+    SpiceAudioClass *audio_class = SPICE_AUDIO_CLASS(klass);
+
+    audio_class->connect_channel = connect_channel;
 
     gobject_class->finalize = spice_gstaudio_finalize;
     gobject_class->dispose = spice_gstaudio_dispose;
 
-    g_type_class_add_private(klass, sizeof(spice_gstaudio));
+    g_type_class_add_private(klass, sizeof(SpiceGstaudioPrivate));
 }
 
 static void record_new_buffer(GstAppSink *appsink, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
     GstMessage *msg;
 
     g_return_if_fail(p != NULL);
@@ -134,8 +135,8 @@ static void record_new_buffer(GstAppSink *appsink, gpointer data)
 
 static void record_stop(SpiceRecordChannel *channel, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
 
     SPICE_DEBUG("%s", __FUNCTION__);
     if (p->record.pipe)
@@ -144,8 +145,8 @@ static void record_stop(SpiceRecordChannel *channel, gpointer data)
 
 static gboolean record_bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
 
     g_return_val_if_fail(p != NULL, FALSE);
 
@@ -176,8 +177,8 @@ static gboolean record_bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
 static void record_start(SpiceRecordChannel *channel, gint format, gint channels,
                          gint frequency, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
 
     g_return_if_fail(p != NULL);
     g_return_if_fail(format == SPICE_AUDIO_FMT_S16);
@@ -216,8 +217,8 @@ static void record_start(SpiceRecordChannel *channel, gint format, gint channels
         p->record.channels = channels;
 
         gst_app_sink_set_emit_signals(GST_APP_SINK(p->record.sink), TRUE);
-        g_signal_connect(p->record.sink, "new-buffer",
-                         G_CALLBACK(record_new_buffer), data);
+        spice_g_signal_connect_object(p->record.sink, "new-buffer",
+                                      G_CALLBACK(record_new_buffer), gstaudio, 0);
 
 lerr:
         g_clear_error(&error);
@@ -232,8 +233,8 @@ lerr:
 static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
                           gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
 
     switch (event) {
     case SPICE_CHANNEL_OPENED:
@@ -256,8 +257,8 @@ static void channel_event(SpiceChannel *channel, SpiceChannelEvent event,
 
 static void playback_stop(SpicePlaybackChannel *channel, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
 
     if (p->playback.pipe)
         gst_element_set_state(p->playback.pipe, GST_STATE_READY);
@@ -269,8 +270,8 @@ static void playback_stop(SpicePlaybackChannel *channel, gpointer data)
 
 static gboolean update_mmtime_timeout_cb(gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
     GstQuery *q;
 
     q = gst_query_new_latency();
@@ -288,8 +289,8 @@ static gboolean update_mmtime_timeout_cb(gpointer data)
 static void playback_start(SpicePlaybackChannel *channel, gint format, gint channels,
                            gint frequency, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
 
     g_return_if_fail(p != NULL);
     g_return_if_fail(format == SPICE_AUDIO_FMT_S16);
@@ -341,8 +342,8 @@ static void playback_data(SpicePlaybackChannel *channel,
                           gpointer *audio, gint size,
                           gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
     GstBuffer *buf;
 
     g_return_if_fail(p != NULL);
@@ -356,11 +357,11 @@ static void playback_data(SpicePlaybackChannel *channel,
 
 static void playback_volume_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
+    SpiceGstaudio *gstaudio = data;
     GstElement *e;
     guint16 *volume;
     guint nchannels;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
     gdouble vol;
 
     if (!p->playback.sink)
@@ -390,8 +391,8 @@ static void playback_volume_changed(GObject *object, GParamSpec *pspec, gpointer
 
 static void playback_mute_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
     GstElement *e;
     gboolean mute;
 
@@ -414,8 +415,8 @@ static void playback_mute_changed(GObject *object, GParamSpec *pspec, gpointer d
 
 static void record_volume_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
     GstElement *e;
     guint16 *volume;
     guint nchannels;
@@ -451,8 +452,8 @@ static void record_volume_changed(GObject *object, GParamSpec *pspec, gpointer d
 
 static void record_mute_changed(GObject *object, GParamSpec *pspec, gpointer data)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = data;
+    SpiceGstaudioPrivate *p = gstaudio->priv;
     GstElement *e;
     gboolean mute;
 
@@ -474,65 +475,62 @@ static void record_mute_changed(GObject *object, GParamSpec *pspec, gpointer dat
     g_object_unref(e);
 }
 
-static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
+static gboolean connect_channel(SpiceAudio *audio, SpiceChannel *channel)
 {
-    SpiceGstAudio *gstaudio = data;
-    spice_gstaudio *p = gstaudio->priv;
+    SpiceGstaudio *gstaudio = SPICE_GSTAUDIO(audio);
+    SpiceGstaudioPrivate *p = gstaudio->priv;
 
     if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
-        g_return_if_fail(p->pchannel == NULL);
+        g_return_val_if_fail(p->pchannel == NULL, FALSE);
+
         p->pchannel = g_object_ref(channel);
-        g_signal_connect(channel, "playback-start",
-                         G_CALLBACK(playback_start), gstaudio);
-        g_signal_connect(channel, "playback-data",
-                         G_CALLBACK(playback_data), gstaudio);
-        g_signal_connect(channel, "playback-stop",
-                         G_CALLBACK(playback_stop), gstaudio);
-        g_signal_connect(channel, "channel-event",
-                         G_CALLBACK(channel_event), gstaudio);
-        g_signal_connect(channel, "notify::volume",
-                         G_CALLBACK(playback_volume_changed), gstaudio);
-        g_signal_connect(channel, "notify::mute",
-                         G_CALLBACK(playback_mute_changed), gstaudio);
-        spice_channel_connect(channel);
+        spice_g_signal_connect_object(channel, "playback-start",
+                                      G_CALLBACK(playback_start), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "playback-data",
+                                      G_CALLBACK(playback_data), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "playback-stop",
+                                      G_CALLBACK(playback_stop), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "channel-event",
+                                      G_CALLBACK(channel_event), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "notify::volume",
+                                      G_CALLBACK(playback_volume_changed), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "notify::mute",
+                                      G_CALLBACK(playback_mute_changed), gstaudio, 0);
+
+        return TRUE;
     }
 
     if (SPICE_IS_RECORD_CHANNEL(channel)) {
-        g_return_if_fail(p->rchannel == NULL);
+        g_return_val_if_fail(p->rchannel == NULL, FALSE);
+
         p->rchannel = g_object_ref(channel);
-        g_signal_connect(channel, "record-start",
-                         G_CALLBACK(record_start), gstaudio);
-        g_signal_connect(channel, "record-stop",
-                         G_CALLBACK(record_stop), gstaudio);
-        g_signal_connect(channel, "channel-event",
-                         G_CALLBACK(channel_event), gstaudio);
-        g_signal_connect(channel, "notify::volume",
-                         G_CALLBACK(record_volume_changed), gstaudio);
-        g_signal_connect(channel, "notify::mute",
-                         G_CALLBACK(record_mute_changed), gstaudio);
-        spice_channel_connect(channel);
+        spice_g_signal_connect_object(channel, "record-start",
+                                      G_CALLBACK(record_start), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "record-stop",
+                                      G_CALLBACK(record_stop), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "channel-event",
+                                      G_CALLBACK(channel_event), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "notify::volume",
+                                      G_CALLBACK(record_volume_changed), gstaudio, 0);
+        spice_g_signal_connect_object(channel, "notify::mute",
+                                      G_CALLBACK(record_mute_changed), gstaudio, 0);
+
+        return TRUE;
     }
+
+    return FALSE;
 }
 
-SpiceGstAudio *spice_gstaudio_new(SpiceSession *session, GMainContext *context,
+SpiceGstaudio *spice_gstaudio_new(SpiceSession *session, GMainContext *context,
                                   const char *name)
 {
-    SpiceGstAudio *gstaudio;
-    spice_gstaudio *p;
-    GList *list;
+    SpiceGstaudio *gstaudio;
 
     gst_init(NULL, NULL);
-    gstaudio = g_object_new(SPICE_TYPE_GSTAUDIO, NULL);
-    p = SPICE_GSTAUDIO_GET_PRIVATE(gstaudio);
-    p->session = g_object_ref(session);
-
-    g_signal_connect(session, "channel-new",
-                     G_CALLBACK(channel_new), gstaudio);
-    list = spice_session_get_channels(session);
-    for (list = g_list_first(list); list != NULL; list = g_list_next(list)) {
-        channel_new(session, list->data, (gpointer)gstaudio);
-    }
-    g_list_free(list);
+    gstaudio = g_object_new(SPICE_TYPE_GSTAUDIO,
+                            "session", session,
+                            "main-context", context,
+                            NULL);
 
     return gstaudio;
 }
