@@ -39,15 +39,17 @@ static gboolean disable_usbredir = FALSE;
 static gint cache_size = 0;
 static gint glz_window_size = 0;
 
+G_GNUC_NORETURN
 static void option_version(void)
 {
     g_print(PACKAGE_STRING "\n");
     exit(0);
 }
 
-static void option_debug(void)
+static gboolean option_debug(void)
 {
     spice_util_set_debug(TRUE);
+    return TRUE;
 }
 
 /**
@@ -155,5 +157,5 @@ void spice_set_session_option(SpiceSession *session)
     if (cache_size)
         g_object_set(session, "cache-size", cache_size, NULL);
     if (glz_window_size)
-        g_object_set(session, "glz_window_size", glz_window_size, NULL);
+        g_object_set(session, "glz-window-size", glz_window_size, NULL);
 }
