@@ -92,6 +92,8 @@ gboolean spice_session_connect(SpiceSession *session);
 gboolean spice_session_open_fd(SpiceSession *session, int fd);
 void spice_session_disconnect(SpiceSession *session);
 GList *spice_session_get_channels(SpiceSession *session);
+gboolean spice_session_has_channel_type(SpiceSession *session, gint type);
+gboolean spice_session_get_read_only(SpiceSession *session);
 
 G_END_DECLS
 
