@@ -20,7 +20,9 @@
 # include "config.h"
 #endif
 #include <glib-object.h>
+#include "spice-util-priv.h"
 #include "spice-util.h"
+#include "spice-util-priv.h"
 
 /**
  * SECTION:spice-util
@@ -76,6 +78,16 @@ gboolean spice_strv_contains(const GStrv strv, const gchar *str)
     return FALSE;
 }
 
+G_GNUC_INTERNAL
+gchar* spice_uuid_to_string(const guint8 uuid[16])
+{
+    return g_strdup_printf(UUID_FMT, uuid[0], uuid[1],
+                           uuid[2], uuid[3], uuid[4], uuid[5],
+                           uuid[6], uuid[7], uuid[8], uuid[9],
+                           uuid[10], uuid[11], uuid[12], uuid[13],
+                           uuid[14], uuid[15]);
+}
+
 typedef struct {
     GObject *instance;
     GObject *observer;
@@ -192,3 +204,9 @@ gulong spice_g_signal_connect_object (gpointer instance,
 
     return ctx->handler_id;
 }
+
+G_GNUC_INTERNAL
+const gchar* spice_yes_no(gboolean value)
+{
+    return value ? "yes" : "no";
+}
