@@ -101,6 +101,7 @@ enum {
     PROP_AUTO_CLIPBOARD,
     PROP_SCALING,
     PROP_DISABLE_INPUTS,
+    PROP_ZOOM_LEVEL
 };
 
 /* Signals */
@@ -123,7 +124,7 @@ static void try_keyboard_ungrab(SpiceDisplay *display);
 static void update_mouse_grab(SpiceDisplay *display);
 static void try_mouse_grab(SpiceDisplay *display);
 static void try_mouse_ungrab(SpiceDisplay *display);
-static void recalc_geometry(GtkWidget *widget, gboolean set_display);
+static void recalc_geometry(GtkWidget *widget);
 static void disconnect_main(SpiceDisplay *display);
 static void disconnect_cursor(SpiceDisplay *display);
 static void disconnect_display(SpiceDisplay *display);
@@ -169,12 +170,45 @@ static void spice_display_get_property(GObject    *object,
     case PROP_DISABLE_INPUTS:
         g_value_set_boolean(value, d->disable_inputs);
 	break;
+    case PROP_ZOOM_LEVEL:
+        g_value_set_int(value, d->zoom_level);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
     }
 }
 
+static void scaling_updated(SpiceDisplay *display)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    GdkWindow *window = gtk_widget_get_window(GTK_WIDGET(display));
+
+    recalc_geometry(GTK_WIDGET(display));
+    if (d->ximage && window) { /* if not yet shown */
+        int ww, wh;
+        gdk_drawable_get_size(gtk_widget_get_window(GTK_WIDGET(display)), &ww, &wh);
+        gtk_widget_queue_draw_area(GTK_WIDGET(display), 0, 0, ww, wh);
+    }
+}
+
+static void update_size_request(SpiceDisplay *display)
+{
+    SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    gint reqwidth, reqheight;
+
+    if (d->resize_guest_enable) {
+        reqwidth = 640;
+        reqheight = 480;
+    } else {
+        reqwidth = d->width;
+        reqheight = d->height;
+    }
+
+    gtk_widget_set_size_request(GTK_WIDGET(display), reqwidth, reqheight);
+    recalc_geometry(GTK_WIDGET(display));
+}
+
 static void spice_display_set_property(GObject      *object,
                                        guint         prop_id,
                                        const GValue *value,
@@ -201,23 +235,11 @@ static void spice_display_set_property(GObject      *object,
         break;
     case PROP_RESIZE_GUEST:
         d->resize_guest_enable = g_value_get_boolean(value);
-        if (d->resize_guest_enable) {
-            gtk_widget_set_size_request(GTK_WIDGET(display), 640, 480);
-            recalc_geometry(GTK_WIDGET(display), TRUE);
-        } else {
-            gtk_widget_set_size_request(GTK_WIDGET(display),
-                                        d->width, d->height);
-        }
+        update_size_request(display);
         break;
     case PROP_SCALING:
         d->allow_scaling = g_value_get_boolean(value);
-        recalc_geometry(GTK_WIDGET(display), FALSE);
-        if (d->ximage &&
-            gtk_widget_get_window(GTK_WIDGET(display))) { /* if not yet shown */
-            int ww, wh;
-            gdk_drawable_get_size(gtk_widget_get_window(GTK_WIDGET(display)), &ww, &wh);
-            gtk_widget_queue_draw_area(GTK_WIDGET(display), 0, 0, ww, wh);
-        }
+        scaling_updated(display);
         break;
     case PROP_AUTO_CLIPBOARD:
         g_object_set(d->gtk_session, "auto-clipboard",
@@ -229,6 +251,10 @@ static void spice_display_set_property(GObject      *object,
         update_keyboard_grab(display);
         update_mouse_grab(display);
         break;
+    case PROP_ZOOM_LEVEL:
+        d->zoom_level = g_value_get_int(value);
+        scaling_updated(display);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
@@ -320,6 +346,23 @@ static GdkCursor* get_blank_cursor(void)
     return gdk_cursor_new(GDK_BLANK_CURSOR);
 }
 
+static gboolean grab_broken(SpiceDisplay *self, GdkEventGrabBroken *event,
+                            gpointer user_data G_GNUC_UNUSED)
+{
+    SpiceDisplayPrivate *d = self->priv;
+
+    SPICE_DEBUG("%s (%d)", __FUNCTION__, event->implicit);
+    if (event->keyboard) {
+        d->keyboard_grab_active = false;
+        g_signal_emit(self, signals[SPICE_DISPLAY_KEYBOARD_GRAB], 0, false);
+    } else {
+        d->mouse_grab_active = false;
+        g_signal_emit(self, signals[SPICE_DISPLAY_MOUSE_GRAB], 0, false);
+    }
+
+    return false;
+}
+
 static void spice_display_init(SpiceDisplay *display)
 {
     GtkWidget *widget = GTK_WIDGET(display);
@@ -327,6 +370,7 @@ static void spice_display_init(SpiceDisplay *display)
 
     d = display->priv = SPICE_DISPLAY_GET_PRIVATE(display);
 
+    g_signal_connect(display, "grab-broken-event", G_CALLBACK(grab_broken), NULL);
     gtk_widget_add_events(widget,
                           GDK_STRUCTURE_MASK |
                           GDK_POINTER_MOTION_MASK |
@@ -335,10 +379,11 @@ static void spice_display_init(SpiceDisplay *display)
                           GDK_BUTTON_MOTION_MASK |
                           GDK_ENTER_NOTIFY_MASK |
                           GDK_LEAVE_NOTIFY_MASK |
-                          GDK_KEY_PRESS_MASK);
+                          GDK_KEY_PRESS_MASK |
+                          GDK_SCROLL_MASK);
     gtk_widget_set_double_buffered(widget, false);
     gtk_widget_set_can_focus(widget, true);
-
+    gtk_widget_set_has_window(widget, true);
     d->keycode_map = vnc_display_keymap_gdk2xtkbd_table(&d->keycode_maplen);
     d->grabseq = spice_grab_sequence_new_from_string("Control_L+Alt_L");
     d->activeseq = g_new0(gboolean, d->grabseq->nkeysyms);
@@ -584,7 +629,6 @@ static GdkGrabStatus do_pointer_grab(SpiceDisplay *display)
         ClipCursor(&client_rect);
     }
 #endif
-    gtk_grab_add(GTK_WIDGET(display));
 
 #ifdef GDK_WINDOWING_X11
     if (status == GDK_GRAB_SUCCESS) {
@@ -593,7 +637,7 @@ static GdkGrabStatus do_pointer_grab(SpiceDisplay *display)
         int threshold;
         GdkWindow *w = GDK_WINDOW(gtk_widget_get_window(GTK_WIDGET(display)));
         Display *x_display = GDK_WINDOW_XDISPLAY(w);
-    
+
         XGetPointerControl(x_display, &accel_numerator, &accel_denominator,
                            &threshold);
         XChangePointerControl(x_display, False, False, accel_numerator,
@@ -656,35 +700,19 @@ static void try_mouse_grab(SpiceDisplay *display)
     d->mouse_last_y = -1;
 }
 
-static void mouse_check_edges(GtkWidget *widget, GdkEventMotion *motion)
+static void mouse_wrap(SpiceDisplay *display, GdkEventMotion *motion)
 {
-    SpiceDisplay *display = SPICE_DISPLAY(widget);
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
-    GdkScreen *screen = gtk_widget_get_screen(widget);
-    gint ww, wh;
+    GdkScreen *screen = gtk_widget_get_screen(GTK_WIDGET(display));
 
-    gdk_drawable_get_size(gtk_widget_get_window(widget), &ww, &wh);
-    int x = (int)motion->x;
-    int y = (int)motion->y;
-    int xr = (int)motion->x_root;
-    int yr = (int)motion->y_root;
+    gint xr = gdk_screen_get_width(screen) / 2;
+    gint yr = gdk_screen_get_height(screen) / 2;
 
-    /* from gtk-vnc: In relative mode check to see if client pointer
-     * hit the window edges, and if so move it back by 100px. This is
-     * important because the pointer in the server doesn't correspond
-     * 1-for-1, and so may still be only half way across the
-     * screen. Without this warp, the server pointer would thus appear
-     * to hit an invisible wall */
-    if (x <= 0) xr += 100;
-    if (y <= 0) yr += 100;
-    if (x >= (ww - 1)) xr -= 100;
-    if (y >= (wh - 1)) yr -= 100;
-
-    if (xr != (int)motion->x_root || yr != (int)motion->y_root) {
+    if (xr != (gint)motion->x_root || yr != (gint)motion->y_root) {
         /* FIXME: we try our best to ignore that next pointer move event.. */
         gdk_display_sync(gdk_screen_get_display(screen));
 
-        gdk_display_warp_pointer(gtk_widget_get_display(widget),
+        gdk_display_warp_pointer(gtk_widget_get_display(GTK_WIDGET(display)),
                                  screen, xr, yr);
         d->mouse_last_x = -1;
         d->mouse_last_y = -1;
@@ -710,7 +738,7 @@ static void try_mouse_ungrab(SpiceDisplay *display)
         int threshold;
         GdkWindow *w = GDK_WINDOW(gtk_widget_get_window(GTK_WIDGET(display)));
         Display *x_display = GDK_WINDOW_XDISPLAY(w);
-    
+
         XGetPointerControl(x_display, &accel_numerator, &accel_denominator,
                            &threshold);
         XChangePointerControl(x_display, True, True, accel_numerator,
@@ -733,10 +761,11 @@ static void update_mouse_grab(SpiceDisplay *display)
         try_mouse_ungrab(display);
 }
 
-static void recalc_geometry(GtkWidget *widget, gboolean set_display)
+static void recalc_geometry(GtkWidget *widget)
 {
     SpiceDisplay *display = SPICE_DISPLAY(widget);
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    gdouble zoom = 1.0;
 
     d->mx = 0;
     d->my = 0;
@@ -745,15 +774,15 @@ static void recalc_geometry(GtkWidget *widget, gboolean set_display)
             d->mx = (d->ww - d->width) / 2;
         if (d->wh > d->height)
             d->my = (d->wh - d->height) / 2;
-    }
+    } else
+        zoom = (gdouble)d->zoom_level / 100;
 
-    SPICE_DEBUG("monitors: id %d, guest %dx%d, window %dx%d, offset +%d+%d",
-                d->channel_id, d->width, d->height, d->ww, d->wh, d->mx, d->my);
+    SPICE_DEBUG("monitors: id %d, guest %dx%d, window %dx%d, zoom %g, offset +%d+%d",
+                d->channel_id, d->width, d->height, d->ww, d->wh, zoom, d->mx, d->my);
 
-    if (d->resize_guest_enable && set_display) {
+    if (d->resize_guest_enable)
         spice_main_set_display(d->main, d->channel_id,
-                               0, 0, d->ww, d->wh);
-    }
+                               0, 0, d->ww / zoom, d->wh / zoom);
 }
 
 /* ---------------------------------------------------------------- */
@@ -1127,16 +1156,26 @@ static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion)
         return true;
 
     gdk_drawable_get_size(gtk_widget_get_window(widget), &ww, &wh);
-    if (spicex_is_scaled(display)) {
+    if (spicex_is_scaled(display) && (d->width != ww || d->height != wh)) {
         double sx, sy;
         sx = (double)d->width / (double)ww;
         sy = (double)d->height / (double)wh;
 
-        /* Scaling the desktop, so scale the mouse coords by same
-         * ratio - ceil() seems to be more accurate - not sure
-         * though - scaling is more likely reversible.. */
-        motion->x = ceil(motion->x * sx);
-        motion->y = ceil(motion->y * sy);
+        /* Scaling the desktop, so scale the mouse coords by the same.
+         * Ratio for the rounding used:
+         * 1) We should be able to send mouse coords for all 4 corner pixels
+         * 2) The GdkMotion events are double's, so we've X.Xfrac and Y.Yfrac
+         * 3) The X and Y integer values always go from 0 - (ww/wh - 1)
+         * 4) Note that even if Xfrac = .99999, X will still go from 0 - (ww-1)
+         *    so x will go from 0.99999 - (ww-1).99999
+         * 5) Given 4, the only way to be sure we can always generate 0
+         *    coordinates is to first floor the input coordinates
+         * 6) To avoid rounding errors causing us to be unable to generate
+         *    coordinates for the bottom row / left column of pixels we ceil
+         *    the end result
+         */
+        motion->x = ceil(floor(motion->x) * sx);
+        motion->y = ceil(floor(motion->y) * sy);
     } else {
         motion->x -= d->mx;
         motion->y -= d->my;
@@ -1154,18 +1193,15 @@ static gboolean motion_event(GtkWidget *widget, GdkEventMotion *motion)
         break;
     case SPICE_MOUSE_MODE_SERVER:
         if (d->mouse_grab_active) {
-            if (d->mouse_last_x != -1 &&
-                d->mouse_last_y != -1) {
-                gint dx = motion->x - d->mouse_last_x;
-                gint dy = motion->y - d->mouse_last_y;
+            gint dx = d->mouse_last_x != -1 ? motion->x - d->mouse_last_x : 0;
+            gint dy = d->mouse_last_y != -1 ? motion->y - d->mouse_last_y : 0;
 
-                spice_inputs_motion(d->inputs, dx, dy,
-                                    button_mask_gdk_to_spice(motion->state));
-            }
+            spice_inputs_motion(d->inputs, dx, dy,
+                                button_mask_gdk_to_spice(motion->state));
 
             d->mouse_last_x = motion->x;
             d->mouse_last_y = motion->y;
-            mouse_check_edges(widget, motion);
+            mouse_wrap(display, motion);
         }
         break;
     default:
@@ -1277,7 +1313,7 @@ static gboolean configure_event(GtkWidget *widget, GdkEventConfigure *conf)
     if (conf->width != d->ww  || conf->height != d->wh) {
         d->ww = conf->width;
         d->wh = conf->height;
-        recalc_geometry(widget, TRUE);
+        recalc_geometry(widget);
     }
     return true;
 }
@@ -1417,6 +1453,25 @@ static void spice_display_class_init(SpiceDisplayClass *klass)
                               G_PARAM_CONSTRUCT |
                               G_PARAM_STATIC_STRINGS));
 
+
+    /**
+     * SpiceDisplay:zoom-level:
+     *
+     * Zoom level in percentage, from 10 to 400. Default to 100.
+     * (this option is only supported with cairo backend when scaling
+     * is enabled)
+     *
+     * Since: 0.10
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_ZOOM_LEVEL,
+         g_param_spec_int("zoom-level", "Zoom Level",
+                          "Zoom Level",
+                          10, 400, 100,
+                          G_PARAM_READWRITE |
+                          G_PARAM_CONSTRUCT |
+                          G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceDisplay::mouse-grab:
      * @display: the #SpiceDisplay that emitted the signal
@@ -1503,7 +1558,6 @@ static void primary_create(SpiceChannel *channel, gint format,
 {
     SpiceDisplay *display = data;
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
-    gboolean set_display = FALSE;
 
     d->format = format;
     d->stride = stride;
@@ -1511,14 +1565,9 @@ static void primary_create(SpiceChannel *channel, gint format,
     d->data_origin = d->data = imgdata;
 
     if (d->width != width || d->height != height) {
-        if (d->width != 0 && d->height != 0)
-            set_display = TRUE;
         d->width  = width;
         d->height = height;
-        recalc_geometry(GTK_WIDGET(display), set_display);
-        if (!d->resize_guest_enable) {
-            gtk_widget_set_size_request(GTK_WIDGET(display), width, height);
-        }
+        update_size_request(display);
     }
 }
 
@@ -1573,19 +1622,10 @@ static void cursor_set(SpiceCursorChannel *channel,
 {
     SpiceDisplay *display = data;
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
+    GdkCursor *cursor = NULL;
 
     cursor_invalidate(display);
 
-    if (d->show_cursor) {
-        gdk_cursor_unref(d->show_cursor);
-        d->show_cursor = NULL;
-    }
-
-    if (d->mouse_cursor) {
-        gdk_cursor_unref(d->mouse_cursor);
-        d->mouse_cursor = NULL;
-    }
-
     if (d->mouse_pixbuf) {
         g_object_unref(d->mouse_pixbuf);
         d->mouse_pixbuf = NULL;
@@ -1601,15 +1641,27 @@ static void cursor_set(SpiceCursorChannel *channel,
                                                    (GdkPixbufDestroyNotify)g_free, NULL);
         d->mouse_hotspot.x = hot_x;
         d->mouse_hotspot.y = hot_y;
+        cursor = gdk_cursor_new_from_pixbuf(gtk_widget_get_display(GTK_WIDGET(display)),
+                                            d->mouse_pixbuf, hot_x, hot_y);
+    } else
+        g_warn_if_reached();
 
-        /* gdk_cursor_new_from_pixbuf() will copy pixbuf data on
-           x11/win32/macos. No worries if rgba pointer is freed
-           after. */
-        d->mouse_cursor = gdk_cursor_new_from_pixbuf(gtk_widget_get_display(GTK_WIDGET(display)),
-                                                     d->mouse_pixbuf, hot_x, hot_y);
+    if (d->show_cursor) {
+        /* unhide */
+        gdk_cursor_unref(d->show_cursor);
+        d->show_cursor = NULL;
+        if (d->mouse_mode == SPICE_MOUSE_MODE_SERVER) {
+            /* keep a hidden cursor, will be shown in cursor_move() */
+            d->show_cursor = cursor;
+            return;
+        }
     }
 
+    gdk_cursor_unref(d->mouse_cursor);
+    d->mouse_cursor = cursor;
+
     update_mouse_pointer(display);
+    cursor_invalidate(display);
 }
 
 static void cursor_hide(SpiceCursorChannel *channel, gpointer data)
@@ -1620,6 +1672,7 @@ static void cursor_hide(SpiceCursorChannel *channel, gpointer data)
     if (d->show_cursor != NULL) /* then we are already hidden */
         return;
 
+    cursor_invalidate(display);
     d->show_cursor = d->mouse_cursor;
     d->mouse_cursor = get_blank_cursor();
     update_mouse_pointer(display);
@@ -1632,7 +1685,7 @@ void spice_display_get_scaling(SpiceDisplay *display, double *sx, double *sy)
     int fbw = d->width, fbh = d->height;
     int ww, wh;
 
-    if (!spicex_is_scaled(display)) {
+    if (!spicex_is_scaled(display) || !gtk_widget_get_window(GTK_WIDGET(display))) {
         *sx = 1.0;
         *sy = 1.0;
         return;
@@ -1719,6 +1772,7 @@ static void disconnect_display(SpiceDisplay *display)
                                          display);
     g_signal_handlers_disconnect_by_func(d->display, G_CALLBACK(invalidate),
                                          display);
+    primary_destroy(d->display, display);
     d->display = NULL;
 }
 
@@ -1879,7 +1933,6 @@ void spice_display_mouse_ungrab(SpiceDisplay *display)
  *
  * Deprecated: 0.8: Use spice_gtk_session_copy_to_guest() instead
  **/
-G_GNUC_DEPRECATED_FOR(spice_gtk_session_copy_to_guest)
 void spice_display_copy_to_guest(SpiceDisplay *display)
 {
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
@@ -1897,7 +1950,6 @@ void spice_display_copy_to_guest(SpiceDisplay *display)
  *
  * Deprecated: 0.8: Use spice_gtk_session_paste_from_guest() instead
  **/
-G_GNUC_DEPRECATED_FOR(spice_gtk_session_paste_from_guest)
 void spice_display_paste_from_guest(SpiceDisplay *display)
 {
     SpiceDisplayPrivate *d = SPICE_DISPLAY_GET_PRIVATE(display);
