@@ -40,9 +40,7 @@
 #include "spice-option.h"
 #include "usb-device-widget.h"
 
-/* config */
-static gboolean fullscreen = false;
-static gboolean version = false;
+typedef struct spice_connection spice_connection;
 
 enum {
     STATE_SCROLL_LOCK,
@@ -51,10 +49,18 @@ enum {
     STATE_MAX,
 };
 
-typedef struct spice_window spice_window;
-typedef struct spice_connection spice_connection;
+#define SPICE_TYPE_WINDOW                  (spice_window_get_type ())
+#define SPICE_WINDOW(obj)                  (G_TYPE_CHECK_INSTANCE_CAST ((obj), SPICE_TYPE_WINDOW, SpiceWindow))
+#define SPICE_IS_WINDOW(obj)               (G_TYPE_CHECK_INSTANCE_TYPE ((obj), SPICE_TYPE_WINDOW))
+#define SPICE_WINDOW_CLASS(klass)          (G_TYPE_CHECK_CLASS_CAST ((klass), SPICE_TYPE_WINDOW, SpiceWindowClass))
+#define SPICE_IS_WINDOW_CLASS(klass)       (G_TYPE_CHECK_CLASS_TYPE ((klass), SPICE_TYPE_WINDOW))
+#define SPICE_WINDOW_GET_CLASS(obj)        (G_TYPE_INSTANCE_GET_CLASS ((obj), SPICE_TYPE_WINDOW, SpiceWindowClass))
 
-struct spice_window {
+typedef struct _SpiceWindow SpiceWindow;
+typedef struct _SpiceWindowClass SpiceWindowClass;
+
+struct _SpiceWindow {
+    GObject          object;
     spice_connection *conn;
     int              id;
     GtkWidget        *toplevel, *spice;
@@ -74,18 +80,41 @@ struct spice_window {
     bool             enable_mnemonics_save;
 };
 
+struct _SpiceWindowClass
+{
+  GObjectClass parent_class;
+};
+
+G_DEFINE_TYPE (SpiceWindow, spice_window, G_TYPE_OBJECT);
+
 struct spice_connection {
     SpiceSession     *session;
     SpiceGtkSession  *gtk_session;
-    spice_window     *wins[4];
+    SpiceWindow     *wins[4];
     SpiceAudio       *audio;
-    char             *mouse_state;
-    char             *agent_state;
+    const char       *mouse_state;
+    const char       *agent_state;
     gboolean         agent_connected;
     int              channels;
     int              disconnecting;
 };
 
+static spice_connection *connection_new(void);
+static void connection_connect(spice_connection *conn);
+static void connection_disconnect(spice_connection *conn);
+static void connection_destroy(spice_connection *conn);
+static void resolution_fullscreen(SpiceWindow *win);
+static void resolution_restore(SpiceWindow *win);
+static void usb_connect_failed(GObject               *object,
+                               SpiceUsbDevice        *device,
+                               GError                *error,
+                               gpointer               data);
+static gboolean is_gtk_session_property(const gchar *property);
+
+/* options */
+static gboolean fullscreen = false;
+static gboolean version = false;
+/* globals */
 static GMainLoop     *mainloop = NULL;
 static int           connections = 0;
 static GKeyFile      *keyfile = NULL;
@@ -93,18 +122,6 @@ static GnomeRRScreen *rrscreen = NULL;
 static GnomeRRConfig *rrsaved = NULL;
 static GnomeRRConfig *rrcurrent = NULL;
 
-static spice_connection *connection_new(void);
-static void connection_connect(spice_connection *conn);
-static void connection_disconnect(spice_connection *conn);
-static void connection_destroy(spice_connection *conn);
-static void resolution_fullscreen(struct spice_window *win);
-static void resolution_restore(struct spice_window *win);
-static void usb_connect_failed(GObject               *object,
-                               SpiceUsbDevice        *device,
-                               GError                *error,
-                               gpointer               data);
-static gboolean is_gtk_session_property(const gchar *property);
-
 /* ------------------------------------------------------------------ */
 
 static int ask_user(GtkWidget *parent, char *title, char *message,
@@ -276,7 +293,7 @@ static int connect_dialog(SpiceSession *session)
 
 /* ------------------------------------------------------------------ */
 
-static void update_status_window(struct spice_window *win)
+static void update_status_window(SpiceWindow *win)
 {
     char status[256];
 
@@ -312,26 +329,26 @@ static void menu_cb_connect(GtkAction *action, void *data)
 
 static void menu_cb_close(GtkAction *action, void *data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     connection_disconnect(win->conn);
 }
 
 static void menu_cb_copy(GtkAction *action, void *data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     spice_gtk_session_copy_to_guest(win->conn->gtk_session);
 }
 
 static void menu_cb_paste(GtkAction *action, void *data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     spice_gtk_session_paste_from_guest(win->conn->gtk_session);
 }
 
-static void window_set_fullscreen(struct spice_window *win, gboolean fs)
+static void window_set_fullscreen(SpiceWindow *win, gboolean fs)
 {
     if (fs) {
 #ifdef WIN32
@@ -348,20 +365,20 @@ static void window_set_fullscreen(struct spice_window *win, gboolean fs)
 
 static void menu_cb_fullscreen(GtkAction *action, void *data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     window_set_fullscreen(win, !win->fullscreen);
 }
 
 static void menu_cb_ungrab(GtkAction *action, void *data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     spice_display_mouse_ungrab(SPICE_DISPLAY(win->spice));
 }
 
 #ifdef USE_SMARTCARD
-static void enable_smartcard_actions(spice_window *win, VReader *reader,
+static void enable_smartcard_actions(SpiceWindow *win, VReader *reader,
                                      gboolean can_insert, gboolean can_remove)
 {
     GtkAction *action;
@@ -419,10 +436,15 @@ static void menu_cb_remove_smartcard(GtkAction *action, void *data)
 #endif
 
 #ifdef USE_USBREDIR
+static void remove_cb(GtkContainer *container, GtkWidget *widget, void *data)
+{
+    gtk_window_resize(GTK_WINDOW(data), 1, 1);
+}
+
 static void menu_cb_select_usb_devices(GtkAction *action, void *data)
 {
     GtkWidget *dialog, *area, *usb_device_widget;
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     /* Create the widgets */
     dialog = gtk_dialog_new_with_buttons(
@@ -432,13 +454,20 @@ static void menu_cb_select_usb_devices(GtkAction *action, void *data)
                     GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
                     NULL);
     gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT);
+    gtk_container_set_border_width(GTK_CONTAINER(dialog), 12);
+    gtk_box_set_spacing(GTK_BOX(gtk_bin_get_child(GTK_BIN(dialog))), 12);
+
     area = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
 
     usb_device_widget = spice_usb_device_widget_new(win->conn->session,
                                                     "%s %s");
     g_signal_connect(usb_device_widget, "connect-failed",
                      G_CALLBACK(usb_connect_failed), NULL);
-    gtk_box_pack_start(GTK_BOX(area), usb_device_widget, TRUE, TRUE, 5);
+    gtk_box_pack_start(GTK_BOX(area), usb_device_widget, TRUE, TRUE, 0);
+
+    /* This shrinks the dialog when USB devices are unplugged */
+    g_signal_connect(usb_device_widget, "remove",
+                     G_CALLBACK(remove_cb), dialog);
 
     /* show and run */
     gtk_widget_show_all(dialog);
@@ -449,7 +478,7 @@ static void menu_cb_select_usb_devices(GtkAction *action, void *data)
 
 static void menu_cb_bool_prop(GtkToggleAction *action, gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
     gboolean state = gtk_toggle_action_get_active(action);
     const char *name;
     gpointer object;
@@ -471,7 +500,7 @@ static void menu_cb_conn_bool_prop_changed(GObject    *gobject,
                                            GParamSpec *pspec,
                                            gpointer    user_data)
 {
-    struct spice_window *win = user_data;
+    SpiceWindow *win = user_data;
     const gchar *property = g_param_spec_get_name(pspec);
     GtkAction *toggle;
     gboolean state;
@@ -483,7 +512,7 @@ static void menu_cb_conn_bool_prop_changed(GObject    *gobject,
 
 static void menu_cb_toolbar(GtkToggleAction *action, gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
     gboolean state = gtk_toggle_action_get_active(action);
 
     gtk_widget_set_visible(win->toolbar, state);
@@ -492,7 +521,7 @@ static void menu_cb_toolbar(GtkToggleAction *action, gpointer data)
 
 static void menu_cb_statusbar(GtkToggleAction *action, gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
     gboolean state = gtk_toggle_action_get_active(action);
 
     gtk_widget_set_visible(win->statusbar, state);
@@ -503,12 +532,12 @@ static void menu_cb_about(GtkAction *action, void *data)
 {
     char *comments = _("gtk test client app for the\n"
         "spice remote desktop protocol");
-    static char *copyright = "(c) 2010 Red Hat";
-    static char *website = "http://www.spice-space.org";
-    static char *authors[] = { "Gerd Hoffmann <kraxel@redhat.com>",
+    static const char *copyright = "(c) 2010 Red Hat";
+    static const char *website = "http://www.spice-space.org";
+    static const char *authors[] = { "Gerd Hoffmann <kraxel@redhat.com>",
                                "Marc-André Lureau <marcandre.lureau@redhat.com>",
                                NULL };
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     gtk_show_about_dialog(GTK_WINDOW(win->toplevel),
                           "authors",         authors,
@@ -523,7 +552,7 @@ static void menu_cb_about(GtkAction *action, void *data)
 
 static gboolean delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     connection_disconnect(win->conn);
     return true;
@@ -532,7 +561,7 @@ static gboolean delete_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
 static gboolean window_state_cb(GtkWidget *widget, GdkEventWindowState *event,
                                 gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
     if (event->changed_mask & GDK_WINDOW_STATE_FULLSCREEN) {
         win->fullscreen = event->new_window_state & GDK_WINDOW_STATE_FULLSCREEN;
         if (win->fullscreen) {
@@ -558,7 +587,7 @@ static gboolean window_state_cb(GtkWidget *widget, GdkEventWindowState *event,
 
 static void grab_keys_pressed_cb(GtkWidget *widget, gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     /* since mnemonics are disabled, we leave fullscreen when
        ungrabbing mouse. Perhaps we should have a different handling
@@ -568,7 +597,7 @@ static void grab_keys_pressed_cb(GtkWidget *widget, gpointer data)
 
 static void mouse_grab_cb(GtkWidget *widget, gint grabbed, gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     win->mouse_grabbed = grabbed;
     update_status(win->conn);
@@ -576,7 +605,7 @@ static void mouse_grab_cb(GtkWidget *widget, gint grabbed, gpointer data)
 
 static void keyboard_grab_cb(GtkWidget *widget, gint grabbed, gpointer data)
 {
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
     GtkSettings *settings = gtk_widget_get_settings (widget);
 
     if (grabbed) {
@@ -597,7 +626,7 @@ static void keyboard_grab_cb(GtkWidget *widget, gint grabbed, gpointer data)
     }
 }
 
-static void restore_configuration(struct spice_window *win)
+static void restore_configuration(SpiceWindow *win)
 {
     gboolean state;
     gchar *str;
@@ -614,7 +643,6 @@ static void restore_configuration(struct spice_window *win)
         return;
     }
 
-    g_return_if_fail(nkeys >= 0);
     if (nkeys > 0)
         g_return_if_fail(keys != NULL);
 
@@ -764,9 +792,7 @@ static const char *spice_display_properties[] = {
 
 static const char *spice_gtk_session_properties[] = {
     "auto-clipboard",
-#ifdef USE_USBREDIR
     "auto-usbredir",
-#endif
 };
 
 static const GtkToggleActionEntry tentries[] = {
@@ -795,12 +821,10 @@ static const GtkToggleActionEntry tentries[] = {
         .label       = N_("Automagic clipboard sharing between host and guest"),
         .callback    = G_CALLBACK(menu_cb_bool_prop),
     },{
-#ifdef USE_USBREDIR
         .name        = "auto-usbredir",
         .label       = N_("Auto redirect newly plugged in USB devices"),
         .callback    = G_CALLBACK(menu_cb_bool_prop),
     },{
-#endif
         .name        = "Statusbar",
         .label       = N_("Statusbar"),
         .callback    = G_CALLBACK(menu_cb_statusbar),
@@ -846,9 +870,7 @@ static char ui_xml[] =
 "      <menuitem action='scaling'/>\n"
 "      <menuitem action='disable-inputs'/>\n"
 "      <menuitem action='auto-clipboard'/>\n"
-#ifdef USE_USBREDIR
 "      <menuitem action='auto-usbredir'/>\n"
-#endif
 "    </menu>\n"
 "    <menu action='HelpMenu'>\n"
 "      <menuitem action='About'/>\n"
@@ -1011,7 +1033,7 @@ end:
     g_clear_error(&error);
 }
 
-static void resolution_fullscreen(struct spice_window *win)
+static void resolution_fullscreen(SpiceWindow *win)
 {
     GnomeRROutputInfo *output;
     int x, y, width, height;
@@ -1041,7 +1063,7 @@ static void resolution_fullscreen(struct spice_window *win)
 #endif
 }
 
-static void resolution_restore(struct spice_window *win)
+static void resolution_restore(SpiceWindow *win)
 {
     GnomeRROutputInfo *output, *saved;
     int x, y, width, height;
@@ -1092,7 +1114,7 @@ static gboolean configure_event_cb(GtkWidget         *widget,
                                    gpointer           data)
 {
     gboolean resize_guest;
-    struct spice_window *win = data;
+    SpiceWindow *win = data;
 
     g_return_val_if_fail(win != NULL, FALSE);
     g_return_val_if_fail(win->conn != NULL, FALSE);
@@ -1109,10 +1131,20 @@ static gboolean configure_event_cb(GtkWidget         *widget,
     return FALSE;
 }
 
-static spice_window *create_spice_window(spice_connection *conn, int id, SpiceChannel *channel)
+static void
+spice_window_class_init (SpiceWindowClass *klass)
+{
+}
+
+static void
+spice_window_init (SpiceWindow *self)
+{
+}
+
+static SpiceWindow *create_spice_window(spice_connection *conn, int id, SpiceChannel *channel)
 {
     char title[32];
-    struct spice_window *win;
+    SpiceWindow *win;
     GtkAction *toggle;
     gboolean state;
     GtkWidget *vbox, *frame;
@@ -1120,7 +1152,7 @@ static spice_window *create_spice_window(spice_connection *conn, int id, SpiceCh
     int i;
     SpiceGrabSequence *seq;
 
-    win = g_new0(struct spice_window, 1);
+    win = g_object_new(SPICE_TYPE_WINDOW, NULL);
     win->id = id;
     win->conn = conn;
     win->display_channel = channel;
@@ -1186,7 +1218,11 @@ static spice_window *create_spice_window(spice_connection *conn, int id, SpiceCh
                      G_CALLBACK(grab_keys_pressed_cb), win);
 
     /* status line */
+#if GTK_CHECK_VERSION(3,0,0)
+    win->statusbar = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 1);
+#else
     win->statusbar = gtk_hbox_new(FALSE, 1);
+#endif
 
     win->status = gtk_label_new("status line");
     gtk_misc_set_alignment(GTK_MISC(win->status), 0, 0.5);
@@ -1206,7 +1242,11 @@ static spice_window *create_spice_window(spice_connection *conn, int id, SpiceCh
     }
 
     /* Make a vbox and put stuff in */
+#if GTK_CHECK_VERSION(3,0,0)
+    vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 1);
+#else
     vbox = gtk_vbox_new(FALSE, 1);
+#endif
     gtk_container_set_border_width(GTK_CONTAINER(vbox), 0);
     gtk_container_add(GTK_CONTAINER(win->toplevel), vbox);
     gtk_box_pack_start(GTK_BOX(vbox), win->menubar, FALSE, FALSE, 0);
@@ -1240,8 +1280,9 @@ static spice_window *create_spice_window(spice_connection *conn, int id, SpiceCh
 
         snprintf(notify, sizeof(notify), "notify::%s",
                  spice_gtk_session_properties[i]);
-        g_signal_connect(win->conn->gtk_session, notify,
-                         G_CALLBACK(menu_cb_conn_bool_prop_changed), win);
+        spice_g_signal_connect_object(win->conn->gtk_session, notify,
+                                      G_CALLBACK(menu_cb_conn_bool_prop_changed),
+                                      win, 0);
     }
 
     toggle = gtk_action_group_get_action(win->ag, "Toolbar");
@@ -1271,18 +1312,23 @@ static spice_window *create_spice_window(spice_connection *conn, int id, SpiceCh
     }
 #endif
 
+#ifndef USE_USBREDIR
+    GtkAction *usbredir = gtk_action_group_get_action(win->ag, "auto-usbredir");
+    gtk_action_set_visible(usbredir, FALSE);
+#endif
+
     gtk_widget_grab_focus(win->spice);
 
     return win;
 }
 
-static void destroy_spice_window(spice_window *win)
+static void destroy_spice_window(SpiceWindow *win)
 {
     SPICE_DEBUG("destroy window (#%d)", win->id);
     g_object_unref(win->ag);
     g_object_unref(win->ui);
     gtk_widget_destroy(win->toplevel);
-    free(win);
+    g_object_unref(win);
 }
 
 /* ------------------------------------------------------------------ */
@@ -1291,9 +1337,9 @@ static void recent_add(SpiceSession *session)
 {
     GtkRecentManager *recent;
     GtkRecentData meta = {
-        .mime_type    = "application/x-spice",
-        .app_name     = "spicy",
-        .app_exec     = "spicy --uri=%u",
+        .mime_type    = (char*)"application/x-spice",
+        .app_name     = (char*)"spicy",
+        .app_exec     = (char*)"spicy --uri=%u",
     };
     char *uri;
 
@@ -1414,7 +1460,7 @@ static void inputs_modifiers(SpiceChannel *channel, gpointer data)
     }
 }
 
-static void display_mark(SpiceChannel *channel, gint mark, spice_window *win)
+static void display_mark(SpiceChannel *channel, gint mark, SpiceWindow *win)
 {
     g_return_if_fail(win != NULL);
     g_return_if_fail(win->toplevel != NULL);
@@ -1484,10 +1530,8 @@ static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data)
     }
 
     if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
-        if (conn->audio != NULL)
-            return;
         SPICE_DEBUG("new audio channel");
-        conn->audio = spice_audio_new(s, NULL, NULL);
+        conn->audio = spice_audio_get(s, NULL);
     }
 
     if (SPICE_IS_USBREDIR_CHANNEL(channel)) {
@@ -1517,10 +1561,6 @@ static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer dat
 
     if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
         SPICE_DEBUG("zap audio channel");
-        if (conn->audio != NULL) {
-            g_object_unref(conn->audio);
-            conn->audio = NULL;
-        }
     }
 
     if (SPICE_IS_USBREDIR_CHANNEL(channel)) {
@@ -1564,6 +1604,8 @@ static spice_connection *connection_new(void)
     if (manager) {
         g_signal_connect(manager, "auto-connect-failed",
                          G_CALLBACK(usb_connect_failed), NULL);
+        g_signal_connect(manager, "device-error",
+                         G_CALLBACK(usb_connect_failed), NULL);
     }
 
     connections++;
@@ -1678,7 +1720,9 @@ int main(int argc, char *argv[])
     spice_connection *conn;
     gchar *conf_file, *conf;
 
+#if !GLIB_CHECK_VERSION(2,31,18)
     g_thread_init(NULL);
+#endif
     bindtextdomain(GETTEXT_PACKAGE, SPICE_GTK_LOCALEDIR);
     bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
     textdomain(GETTEXT_PACKAGE);
