@@ -1,6 +1,6 @@
 /* -*- Mode: C; c-basic-offset: 4; indent-tabs-mode: nil -*- */
 /*
-   Copyright (C) 2011 Red Hat, Inc.
+   Copyright (C) 2011,2012 Red Hat, Inc.
 
    Red Hat Authors:
    Hans de Goede <hdegoede@redhat.com>
@@ -25,16 +25,15 @@
 
 G_BEGIN_DECLS
 
-#ifdef USE_USBREDIR
-const char *spice_usb_device_manager_libusb_strerror(enum libusb_error error_code);
-#endif
-
 gboolean spice_usb_device_manager_start_event_listening(
     SpiceUsbDeviceManager *manager, GError **err);
 
 void spice_usb_device_manager_stop_event_listening(
     SpiceUsbDeviceManager *manager);
 
+void spice_usb_device_manager_device_error(
+    SpiceUsbDeviceManager *manager, SpiceUsbDevice *device, GError *err);
+
 G_END_DECLS
 
 #endif /* __SPICE_USB_DEVICE_MANAGER_PRIV_H__ */
